/*
 *
 * Modification History:
 *
 * March 30, 1993 by Robin Miller, MILLER W J & ASSOCIATES.
 *	Unlock/lock the modem tty before/after system() calls so
 *	external programs, such as 'qcp', can open the modem port.
 *
 * March 26, 1993 by Robin Miller, MILLER W J & ASSOCIATES.
 *	During abort process, always restore terminal characteristics
 *	of both the local tty and the modem tty since we're exiting.
 *	Previously, it was possible to exit without proper restore.
 *	For non-QNX systems, set "exclusive-use" mode on modem port.
 *	Drain local output before flushing to messages aren't lost.
 *	If modem port can't be opened, report the error msg string.
 *
 * January 8, 1993 by Robin Miller, MILLER W J & ASSOCIATES.
 *	Ported 'qtalk' to DEC OSF/1 Operating System.
 *	o  made 8-bit safe (to avoid sign extension, etc).
 *	o  added functions to setup terminal attributes.
 *	o  use select() system call instead of proxies.
 *	o  added '-b baud_rate' to set the modem speed.
 *	   added '-f flow_control' to set this attribute.
 *	   added '-t parity_type' to set this attribute.
 *
 * NOTE: Unlike QNX, the speed and other attributes, cannot be set
 *	 via 'stty', since the terminal driver always sets up defaults
 *	 on the 1st tty open. which resets the speed, parity, etc.
 *
 * BTW: If logging into a DEC OSF/1 system, you'll also need to create
 *	the following entries in /etc/gettydefs:
 *
 * 19200# B19200 HUPCL CS8 CLOCAL # B19200 SANE IXANY TAB3 CLOCAL #login: #19200
 * M19200# B19200 HUPCL CS8 # B19200 SANE IXANY TAB3 HUPCL #login: #M19200
 *
 * The tty defaults setup on the "first" open are (for my reference):
 *
 * #define TTYDEF_IFLAG	(BRKINT | ICRNL | IMAXBEL | IXON | IXANY)
 * #define TTYDEF_OFLAG	(OPOST | ONLCR)
 * #define TTYDEF_LFLAG	(ECHO | ICANON | ISIG | IEXTEN | ECHOE|ECHOKE|ECHOCTL)
 * #define TTYDEF_CFLAG	(CREAD | CS8 | HUPCL)
 * #define TTYDEF_SPEED	(B9600)
 *
 *	What a pain in the ass, eh?  No wonder all UNIX programs have to
 *	be over complicated with all this junk (can't setup with 'stty').
 */
#if defined(_QNX_SOURCE)
#define STEAL_ALLOWED "S"

#ifdef __USAGE
%C	[options]*
Options:
 -c <cmd>        Command char.
 -d <del>        Delete char.
 -e              Local echo on.
 -E              Suppress FF codes (qnxs).
 -i              Input flow control on.
 -k <kick>       Kick char.
 -l <logfile>    Path to log file.
 -m <port>       Device to use.
 -o <options>    Transfer protocol options.
 -p <pause>      Pause char.
 -P              Strip parity.
 -s <system>     System to call.
Note:
 For HELP, press Ctrl-A followed by ? within Qtalk.
#endif

#endif /* defined(_QNX_SOURCE) */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#if defined(_QNX_SOURCE)
#  include <sys/dev.h>
#  include <sys/kernel.h>
#  include <sys/proxy.h>
#else /* !defined(_QNX_SOURCE) */
#  include <sys/ioctl.h>
#  include <sys/time.h>
#endif /* defined(_QNX_SOURCE) */
#include <sys/wait.h>
#include <termios.h>
#include <malloc.h>
#include <signal.h>
#include <string.h>
#include <setjmp.h>
#include <fcntl.h>
#if defined(_QNX_SOURCE)
#  include <process.h>
#endif /* defined(_QNX_SOURCE) */
#include <errno.h>

#if !defined(_OSF_SOURCE)

typedef unsigned char uchar_t;
typedef unsigned int  uint_t;

#endif /* !defined(_OSF_SOURCE) */

#if defined(ultrix)

#define const

typedef int			ssize_t;

#endif /* defined(ultrix) */

/*
 * Forward References:
 */
void	a_abrt(int err, char *fn, int fd);
ssize_t	a_read(int fildes, void *buf, size_t nbyte);
ssize_t	a_write(int fildes, const void *buf, size_t nbyte);
int	a_dev_ischars(int fd);
int	a_tcdrain(int fildes);
uchar_t	*get_name( uchar_t *prompt );
void	gotsig( int sig );
void	mdm_handler( uchar_t *bp, int size, int *armed );
int	put_console( uchar_t *buf, int n );
int	put_logfile( uchar_t *buf, int n );
int	dial( uchar_t *system );
void	kbd_handler( uchar_t *bp, int size );
void	command(void);
void	mdm_write( int upload_flag );


#if defined(_QNX_SOURCE)

#define old_mode()	{a_dev_mode(kbd, okbd_mode, _DEV_MODES);\
					 a_dev_mode(mdm, omdm_mode, _DEV_MODES);}

#define old_kbdmode()	{a_dev_mode(kbd, okbd_mode, _DEV_MODES);}

#define new_mode()	{a_dev_mode(kbd,  kbd_mode, _DEV_MODES);\
					 a_dev_mode(mdm,  mdm_mode, _DEV_MODES);}
#define new_kbdmode()	{a_dev_mode(kbd,  kbd_mode, _DEV_MODES);}

int	a_dev_read(int fd, void *buf, unsigned n, unsigned min, unsigned time,
               unsigned timeout, pid_t proxy, int *armed);
int	a_dev_arm(int fd, pid_t proxy, unsigned events);
uint_t	a_dev_mode(int fd, unsigned mode, unsigned mask);

#else /* !defined(_QNX_SOURCE) */

extern int	getopt(int, char *[],char *);
extern char	*optarg;
extern int	optind, opterr;

#define match(s1,s2)	(strcmp(s1,s2)==0)

#define atoh(sptr)	strtol((char *)sptr, (char **)NULL, 16)

#define old_mode()	{restore_tty(kbd, &local_tty);\
				restore_tty(mdm, &modem_tty);}
#define old_kbdmode()	{restore_tty(kbd, &local_tty);}

#define new_mode()	{setup_local(kbd, &local_tty);\
				setup_modem(mdm, &modem_tty);}
#define new_kbdmode()	{setup_local(kbd, &local_tty);}

#define LOCK_TTY(fd)	if (!debug_flag) {			\
			    if (ioctl (fd, TIOCEXCL, 0) < 0) {	\
				perror ("TIOCEXCL");		\
			    }					\
			}

#define UNLOCK_TTY(fd)	if (!debug_flag) {			\
			    if (ioctl (fd, TIOCNXCL, 0) < 0) {	\
				perror ("TIOCNXCL");		\
			    }					\
			}

enum flow {FLOW_NONE, CTS_RTS, XON_XOFF};

/*
 * Table to Map Baudrate to tty Speed Codes.
 */
struct tty_baud_rate {
	speed_t	usr_speed;		/* User entered speed value.	*/
	speed_t	tty_speed;		/* Parameter for tty driver.	*/
} baud_rate_tbl[] = {
    {	0,	B0	},
    {	50,	B50	},
    {	75,	B75	},
    {	110,	B110	},
    {	134,	B134	},
    {	150,	B150	},
    {	200,	B200	},
    {	300,	B300	},
    {	600,	B600	},
    {	1200,	B1200	},
    {	1800,	B1800	},
    {	2400,	B2400	},
    {	4800,	B4800	},
    {	9600,	B9600	},
    {	19200,	B19200	},
    {	38400,	B38400	}
};
int num_baud_rates = sizeof(baud_rate_tbl) / sizeof(baud_rate_tbl[0]);

char *speed_str = "9600";		/* The baud rate string.	*/
char *flow_str = "xon_xoff";		/* The flow cntrol string.	*/
char *parity_str = "none";		/* The parity type string.	*/
speed_t baud_rate = 0;			/* User specified baud rate.	*/
speed_t default_speed = B9600;		/* Default baud rate is 9600.	*/
unsigned default_parity = 0;		/* Default parity = none.	*/
unsigned default_data_bits = CS8;	/* Default data bits = 8.	*/
enum flow default_flow = XON_XOFF;	/* Terminal flow control type.	*/

struct tty_params {
	int	tty_saved;		/* tty characteristics saved.	*/
	int	tty_ldisc;		/* The saved line discipline.	*/
	struct termios tty_smodes;	/* The saved terminal modes.	*/
	struct termios tty_cmodes;	/* The current terminal modes.	*/
	enum flow tty_flow;		/* The terminal flow control.	*/
	speed_t	tty_speed;		/* The terminal baud rate.	*/
	uint_t	tty_parity;		/* The terminal parity type.	*/
	uint_t	tty_data_bits;		/* The terminal data bits.	*/
} local_tty, modem_tty;

/*
 * Non-QNX O/S Forward References.
 */
int	dev_ischars(int fd);
int	tcdropline(int fd, int hsec);
int	save_tty(int fd, struct tty_params *tp);
int	restore_tty(int fd, struct tty_params *tp);
int	setup_local(int fd, struct tty_params *tp);
int	setup_modem(int fd, struct tty_params *tp);

#endif /* defined(_QNX_SOURCE) */

/*
 * Generic return codes for library functions/system calls.
 */
#define SUCCESS				0
#define FAILURE				-1
#define WARNING				1

/*
 * Return code for boolean functions.
 */
#define TRUE				1
#define FALSE				0

#define VERSION				"[4.1X]"
#define MDM_VERSION			"q10"
#define BUFFER_SIZE			4000
#define DEFAULT_CMD_CHAR		0x01
#define ENQ				0x05
#define SYN				0x16
#define ESC				0x1B
#define RUB				0x7F
#if defined(_QNX_SOURCE)
#  define DIAL_DIR			"/etc/config/dial.dir"
#else /* !defined(QNX_SOURCE) */
#  define DIAL_DIR			"/usr/lib/dial.dir"
#endif /* defined(_QNX_SOURCE) */
#define LINE_LENGTH			256
#define OPT_LENGTH			LINE_LENGTH

int					mdm = -1, scr, kbd, log_file = -1;
int					mdmflag, kbdflag, qnxs;
pid_t					pid;
unsigned				mproxy = 0, kproxy = 1, vmproxy = 0, vkproxy = 1;

sigjmp_buf				jmpbuf;
unsigned				kbd_mode, mdm_mode, okbd_mode, omdm_mode;

#if defined(_QNX_SOURCE)
struct	_dev_info_entry	kbd_entry, mdm_entry;
#endif /* defined(_QNX_SOURCE) */

uchar_t			*mname,
			mtbuf[32],
			buf[ BUFFER_SIZE ],
			buf2[ BUFFER_SIZE ],
			*sys,
			cmd_char,
			del_char,
			pause_char,
			kick_char,
			protocol[ LINE_LENGTH + 1 ],
			transfer_options[ OPT_LENGTH + 1 ],
			line[ LINE_LENGTH + 1 ];
int			num_packets;
long			num_chars;
int			show_stats,
			has_timeouts,
			local_echo,
			strip_parity,
			numbrks;

char	*cmdname = "qtalk";
int	debug_flag = FALSE;
int	modem_flag = FALSE;

/*-------------------------------------------------- IO COVER FNS
        They abort when they get 'significant' errors.
*/

void
a_abrt(int err, char *fn, int fd)
{
	fprintf(stderr,"\r\nqtalk: (%s) %s\r\n",fn,strerror(err));
	/*
	 * Restore the previous terminal characteristics.
	 */
#if defined(_QNX_SOURCE)
	if (mdm!=-1 && fd!=mdm) tcdropline(mdm,500);
	if (kbd!=-1 && fd!=kbd) dev_mode(kbd, okbd_mode, _DEV_MODES);
	if (mdm!=-1 && fd!=mdm) dev_mode(mdm, omdm_mode, _DEV_MODES);
	if (err==ESRCH || err==EBADF) {
		if (fd==mdm) fprintf(stderr,"qtalk: May have lost network VC to modem device.\r\n");
	}
#else /* !defined(_QNX_SOURCE) */
	(void) restore_tty (kbd, &local_tty);
	if (mdm != -1) (void) tcdropline(mdm, 500);
	(void) restore_tty (mdm, &modem_tty);
#endif /* defined(_QNX_SOURCE) */

	exit (EXIT_FAILURE);
}

ssize_t
a_read(int fildes, void *buf, size_t nbyte)
{
	ssize_t rc;	

/*	printf("a_read(int fildes = %d, void *buf = %#lx, size_t nbyte = %d\n", fildes, buf, nbyte);	*/
	if ((rc = read(fildes,buf,nbyte))==-1) {
		switch (errno) {
			case EINTR:
			case EIO:		/* Old MacDonald had a farm, EI EIO... */
				break;
			default:		/* Old MacDonald bought the farm... */
				a_abrt(errno,"read",fildes);
		}
	}	
	return rc;
}

ssize_t
a_write(int fildes, const void *buf, size_t nbyte)
{
	ssize_t rc;

	if ((rc = write(fildes,buf,nbyte))==-1) {
		switch (errno) {
			case EINTR:
			case EIO:
#if defined(ultrix)
				rc = -1;
#else /* !defined(ultrix) */
				rc=-1;
#endif /* defined(ultrix) */
				break;
			default:
				a_abrt(errno,"write",fildes);
		}
	}

	return rc;
}

int
a_dev_ischars(int fd)
{
	int rc;

	if ((rc = dev_ischars(fd))==-1)
		a_abrt(errno,"dev_ischars",fd);
	return rc;                      
}


#if defined(_QNX_SOURCE)

int
a_dev_arm(int fd, pid_t proxy, unsigned events)
{
	int rc;

	if ((rc = dev_arm(fd,proxy,events))==-1) 
		a_abrt(errno,"dev_arm",fd);

	return rc;
}

unsigned
a_dev_mode(int fd, unsigned mode, unsigned mask)
{
	unsigned rc;

	if ((rc = dev_mode(fd,mode,mask))==-1) 
		a_abrt(errno,"dev_mode",fd);
	return rc;
}

int
a_dev_read(int fd, void *buf, unsigned n, unsigned min, unsigned time,
               unsigned timeout, pid_t proxy, int *armed)
{
	int rc;

	if ((rc = dev_read(fd,buf,n,min,time,timeout,proxy,armed))==-1) {
		switch (errno) {
			case EINTR:
			case EIO:
				break;
			default:
				a_abrt(errno,"dev_read",fd);
		}
	}
	return rc;
}

#endif /* defined(_QNX_SOURCE) */

int
a_tcdrain(int fildes)
{
	int rc;

	if ((rc = tcdrain(fildes))==-1) {
		switch(errno) {
			case EINTR:
				break;
			default:
				a_abrt(errno,"tcdrain",fildes);
		}
	}
	return rc;
}

/* end of IO cover fns ------------------------------------------------ */				


uchar_t *
get_name( uchar_t *prompt )
{
	static uchar_t	name[ LINE_LENGTH + 1 ];
	int			i;

	old_kbdmode();
	a_write( kbd, prompt, strlen( (char *)prompt ) );
	(void)fgets( (char *)name, LINE_LENGTH, stdin );
	new_kbdmode();

	i = strlen((char *)name)-1;
	if ( name[i] == '\n' )
		name[i] = '\0';

	return( name );
}

void
gotsig( int sig )
{
	if ( sig == SIGINT ) {
		abort();
		tcsendbreak( mdm, 300 );
		siglongjmp( jmpbuf, 0 );
	}
	old_mode();
	exit( sig );
}

void
mdm_handler( uchar_t *bp, int size, int *armed )
{
	static int		numsyns,
				esc_seen;
	uchar_t			*p;
	int			n;

	put_console( bp, size);
	/*
	 * post-process the line
	 */
	p = bp;
	for(n = 0; n < size; ++n) {
		/*
		 * Respond to 2 consecutively received SYN characters
		 */
		if(*p == SYN) {
			if(++numsyns == 2 ) {
				/*
				 *	After 2 in a row, invoke QCP.
				 */

				/* turn off modem proxies */
#if defined(_QNX_SOURCE)
				a_dev_arm(mdm,-1,_DEV_EVENT_INPUT);

				/* flush any pending modem proxies */
				while (Creceive(mproxy,bp,0)!=-1);
#endif /* defined(_QNX_SOURCE) */

				/* read and discard any pending modem data */
				while (n=a_dev_ischars(mdm)) {
					if (n==-1) break;
					/* have BUFFER_SIZE to play with */
					a_read(mdm,bp,n>BUFFER_SIZE?BUFFER_SIZE:n);
				}
				
				(void)sprintf( (char *)line, "%s %s re %s", protocol, mname, transfer_options );
#if defined(_QNX_SOURCE)
				old_mode();
#else /* !defined(_QNX_SOURCE) */
				/*
				 * We can't reset the modem attributes on DEC OSF/1,
				 * otherwise we lose the baudrate we originally setup.
				 */
				old_kbdmode();
				UNLOCK_TTY(mdm);
#endif /* defined(_QNX_SOURCE) */
				system( (char *)line );
				new_mode();
				numsyns = 0;

				/* turn modem proxies back on */
#if defined(_QNX_SOURCE)
				a_dev_arm(mdm,vmproxy,_DEV_EVENT_INPUT);
#else /* !defined(_QNX_SOURCE) */
				LOCK_TTY(mdm);
#endif /* defined(_QNX_SOURCE) */

				*armed = 0;	/* ready to go back to receive */

				errno=0;
				break;	/* ignore the rest of the line */
			}
		} else numsyns = 0;
		/*
		 * Write back ID after <Esc><Enq>
		 */
		if(esc_seen && *p == ENQ) {
			a_write( mdm, MDM_VERSION, 3 );
		}
		if(*p == ESC) esc_seen = 1;
		else esc_seen = 0;
		++p;
	}
}

int
put_console( uchar_t *buf, int n )
{
	int	 i;
	uchar_t *p = buf;

	if (strip_parity)	for(i=0; i<n; ++p,++i) *p &= 0x7F;

	if (qnxs) {
		uchar_t *buf2=buf;	/* current print position */
		uchar_t *replace_with;
		int  numtoprint=0;

		for (i=0;i<n;++i,++numtoprint) {
			switch(*p) {
				case 0xa6:	replace_with = (uchar_t *)"c"; break; /* right arrow */
				case 0xa4:	replace_with = (uchar_t *)"d"; break; /* left arrow */
				case 0xa1:	replace_with = (uchar_t *)"a"; break; /* up arrow */
				case 0xa9:	replace_with = (uchar_t *)"b"; break; /* down arrow */
				case 0xa0:	replace_with = (uchar_t *)"H"; break; /* home key */
				default: replace_with = NULL; break;
			}

			if (replace_with!=NULL) {
				a_write(kbd,buf2,numtoprint);	/* write line to this point */
				if (log_file != -1) put_logfile( buf2, numtoprint );
				buf2+=numtoprint+1;			/* increment 'write from' position */
				a_write(kbd,replace_with,2);	/* write replacement for ctrl sequence */
				if (log_file != -1) put_logfile( replace_with, 2 );
#if defined(ultrix)
				numtoprint = -1;	/* will be inc'ed to 0 as we loop */
#else /* !defined(ultrix) */
				numtoprint=-1;		/* will be inc'ed to 0 as we loop */
#endif /* defined(ultrix) */
			}
		}
		if (numtoprint) a_write(kbd,buf2,numtoprint);
		if (log_file != -1) put_logfile( buf2, numtoprint );
	} else {
		a_write( kbd, buf, n );
		if (log_file != -1) put_logfile( buf, n );
	}

	return( n );
}

int
put_logfile( uchar_t *buf, int n )
{
	register uchar_t *p1,
			 *p2,
			 *endp;
	register int 	 count = 0;

	for( p1 = p2 = buf, endp = p1 + n; p1 < endp; ++p1 ) {
		if(*p1 != '\r') {				/* Strip CRs */
			*p2++ = *p1;
			++count;
		}
	}

	write( log_file, buf, count );

	return  n;
}

int
dial( uchar_t *system )
{
	FILE	*fp;
	uchar_t	*p;
	int	n;

	if (!(fp=fopen(DIAL_DIR,"r")))	return 0;

	if ( strcmp( (char *)system, "?" ) == 0 ) {
		while( 1 ) {
			p = (uchar_t *)fgets( (char *)line, LINE_LENGTH, fp );

			if(p == NULL) break;
			n = strlen((char *)p);

			if(n<1)  break;
			line[n-1] = '\0';
			fprintf( stderr, "%s\r\n", line );
		}
		fclose( fp );
		return 2;
	} else while( 1 ) {
		p = (uchar_t *)fgets( (char *)line, LINE_LENGTH, fp );
		if(p == NULL) break;
		n = strlen( (char *)p );
		if( n < 1) continue;
		line[n - 1] = '\0';
		for(p=line; *p; ++p) if (*p==' '||*p=='\t') break;

		if (*p == '\0') continue;		/*	Check for a dialing string	*/

		*p++ = '\0';
		if ( strcmp( (char *)line, (char *)system ) == 0 ) {
			/*
			 *	Skip leading whitespace before the dialing string
			 */
			while(*p && (*p==' ' || *p=='\t')) ++p;

			if (*p == '\0') break;	/*	Check for null dialing string	*/

			a_write( mdm, p, strlen( (char *)p ) );
			a_write( mdm, "\r", 1 );
			fclose( fp );
			return 1;
		}
	} /* else while (1) */

	fprintf( stderr, "Unable to dial '%s': no dialing entry\r\n", system );
	fclose( fp );
	return 0;
}

void
kbd_handler( uchar_t *bp, int size )
{
	uchar_t	*p, *p1;
	unsigned char c;
	int		n=0;
	int command_flag = 0;
	int i;

	p = p1 = bp;

	for(i = 0; i < size; ++i) {
		c = *p++;
		if(c == cmd_char) {
			++command_flag;
			break; /* Ignore the rest of the data */
		}
		if (qnxs && c==0xFF) continue;
		if (del_char && c==RUB)	c=del_char;
		*p1++ = c;
		++n;
	}

	if(n > 0) {
		a_write( mdm, bp, n );
		if (local_echo) put_console( bp, n );
	}

	if(command_flag) command();

}


void
command(void)
{
	register uchar_t	*p;
	uchar_t			 c;

	printf( "\r\nQtalk %s\r\n",VERSION);
	printf( "Local echo %s\r\n",local_echo?"enabled":"disabled");
	printf( "Logging    %s\r\n",log_file!=-1?"enabled":"disabled");
	printf( "Parity bit %s\r\n",strip_parity?"stripped":"displayed");

again:
	printf( "\r\nCommand(?): ");
	fflush( stdout );
	read( kbd, &c, 1 );
	if (isalnum(c) || c=='?') putc( c, stdout );

	printf( "\r\n" );
	switch(tolower(c)) {
		case 'b':	tcsendbreak( mdm, 300 );	break;

		case 'c':	getcwd( (char *)line, LINE_LENGTH );
				printf( "Current Directory is: %s\r\n", line);
				p = get_name( (uchar_t *)"Directory? " );
				if (strlen((char *)p) > 0) {
					if ( chdir( (char *)p ) != 0 )
						printf( "Unable to change to directory %s\r\n", p );
				}
				break;

		case 'd':	do {
					p = get_name((uchar_t *)"System? (? for list) ");
					if (*p=='\0') break;
				} while( dial( p ) == 2 );
				break;

		case 'e':	local_echo=!local_echo;
				printf("Local echo %s\r\n",local_echo?"enabled":"disabled");
				break;

		case 'l':	if ( log_file != -1 ) {	/*	toggle log status	*/
					close( log_file );
					log_file = -1;
					printf( "Log file closed\r\n");
				} else log_file=open((char *)get_name((uchar_t *)"Log file? "),O_WRONLY|O_APPEND|O_CREAT,0666);
				break;

		case 'o':	printf("Current transfer options are \"%s\"\r\n",transfer_options );
				p = get_name( (uchar_t *)"Transfer options: " );
				if(strlen((char *)p)>0) strcpy( (char *)transfer_options, (char *)p );
				break;

		case 't':	printf("Current protocol is \"%s\"\r\n",protocol);
				p = get_name( (uchar_t *)"Protocol: " );
				if (strlen((char *)p)>0) strcpy((char *)protocol,(char *)p);
				break;

		case 'p':	strip_parity=!strip_parity;
				printf("Parity bit %s\r\n",strip_parity?"stripped":"displayed");
				break;

		case 'r':	sprintf((char *)line,"%s %s re %s",protocol,mname,transfer_options);
#if defined(_QNX_SOURCE)
				old_mode();
#else /* !defined(_QNX_SOURCE) */
				/*
				 * We can't reset the modem attributes on DEC OSF/1,
				 * otherwise we lose the baudrate we originally setup.
				 */
				old_kbdmode();
				UNLOCK_TTY(mdm);
#endif /* defined(_QNX_SOURCE) */
				system( (char *)line );
				new_mode();
#if !defined(_QNX_SOURCE)
				LOCK_TTY(mdm);
#endif /* !defined(_QNX_SOURCE) */
				break;

		case 's':	sprintf((char *)line,"%s %s se %s %s",protocol,mname,transfer_options,
						 get_name( (uchar_t *)"Send file(s)? " ) );
#if defined(_QNX_SOURCE)
				old_mode();
#else /* !defined(_QNX_SOURCE) */
				old_kbdmode();
				UNLOCK_TTY(mdm);
#endif /* defined(_QNX_SOURCE) */
				system( (char *)line );
				new_mode();
#if !defined(_QNX_SOURCE)
				LOCK_TTY(mdm);
#endif /* !defined(_QNX_SOURCE) */
				break;

		case 'u':	mdm_write( 1 );	break;

		case 'w':	mdm_write( 0 );	break;

		case 'h':	tcdropline(mdm, 500);	break;

		case 'q':	tcdropline( mdm, 500 );	/* fall thru */
		case 'x':	old_mode();
					exit(EXIT_SUCCESS);

		case '2':	if (qnxs) {
					printf("qnxs mode disabled. (Leading FFs will be xmitted)\r\n");
					qnxs=0;
				} else {
					printf("qnxs mode enabled. (Leading FFs will be stripped)\r\n");
					qnxs++;
				}
				break;

		case '!':	p = get_name((uchar_t *)"Enter command: ");
				if(strlen((char *)p) > 0) {
#if defined(_QNX_SOURCE)
					old_mode();
#else /* !defined(_QNX_SOURCE) */
					old_kbdmode();
					UNLOCK_TTY(mdm);
#endif /* defined(_QNX_SOURCE) */
					system( (char *)p );
					new_mode();
#if !defined(_QNX_SOURCE)
					LOCK_TTY(mdm);
#endif /* !defined(_QNX_SOURCE) */
				}
				break;

		default:	if( c == cmd_char ) {
					a_write( mdm, &c, 1 );
					break;
				}
				if (isalnum(c)||c=='?') {
					printf( "%s\r\n%s\r\n%s\r\n%s\r\n%s\r\n%s\r\n%s\r\n%s\r%s\n",
						"b - Send Break             q - Quit & Hangup",
						"c - Change directory       r - Receive a file",
						"d - Dial System            s - Send File",
						"e - Toggle Local Echo      t - Specify transfer protocol",
						"h - Hangup                 u - Upload Text",
						"l - Log File On/Off        w - Write File to Modem",
						"o - Set transfer options   x - Exit w/o Hangup",
						"p - Toggle Parity Filter   ! - Escape to shell",
						"2 - Toggle qnxs mode (To talk to QNX2.x tset qnxs.)"
					);
					goto again;
				}
	}
}

/* this is used only by -u and -w options. What the heck is this for? */

void
mdm_write( int upload_flag )
{
	int		fd, n, i;
	uchar_t		c, kick_seen;

	if ( ( fd = open( (char *)get_name( (uchar_t *)"Upload file? " ), O_RDONLY ) ) == -1 )
		return;

	while( a_read( fd, &c, 1 ) == 1 ) {
		if ( local_echo ) {
			if (c=='\n') put_console((uchar_t *)"\r\n", 2);
			else put_console( &c, 1 );
		}

		if (c=='\n') c='\r';

		a_write( mdm, &c, 1 );

		kick_seen = (upload_flag && !local_echo	|| pause_char && kick_char
					 && c==pause_char)? 0 : 1 ;
		/*
		 *	Only flush if you are about to wait for echo or kick.
		 *  99% of the time a <W>rite won't go through this code.
		 */
		if (!kick_seen) a_tcdrain(mdm);

		while(!kick_seen) {
#ifdef DEVARM
			if ( !mdmflag ) {
				a_dev_arm( mdm, mproxy, _DEV_EVENT_INPUT );
				mdmflag = 1;
			}
			if ( !kbdflag ) {
				a_dev_arm( kbd, kproxy, _DEV_EVENT_INPUT );
				kbdflag = 1;
			}
			pid = Receive( 0, 0, 0 );
#else
			for( ;; ) {
				if ( a_dev_ischars( mdm ) ) {
					pid = mproxy;
					break;
				}
				else if ( a_dev_ischars( kbd ) ) {
					pid = kproxy;
					break;
				}
			}
#endif

			if ( pid == kproxy ) {
				printf( "Upload aborted.\r\n" );
				goto end;
			} else if ( pid == mproxy ) {
				n = a_read( mdm, buf, a_dev_ischars( mdm ) );
				put_console( buf, n );
				if ( pause_char  &&  kick_char  &&  c == pause_char ) {
					for( kick_seen = i = 0; !kick_seen  &&  i < n; ++i ) {
						if (buf[i]==kick_char)	kick_seen = 1;
					}
				} else kick_seen = 1;
			} else {
				fprintf(stderr, "qtalk: Received an unexpected message from pid %d\r\n", pid );
			}
		}
	}

	a_tcdrain( mdm );
#if defined(ultrix)
	printf( "Upload completed\r\012");
#else /* !defined(ultrix) */
	fflush( stdout );
#endif /* defined(ultrix) */
end:
	close( fd );
}

/*
 * main() - Main entry point for 'qtalk' program.
 */
int
main( int argc, char *argv[] )
{
	unsigned			 iflow = 0,
#ifdef STEAL_ALLOWED
					 steal = 0,
#endif
					 opt,
					 n;
int					 armed;
	uchar_t				*cp;
#if !defined(_QNX_SOURCE)
	fd_set	readmask;
	int	nfds, nf;
#endif /* !defined(_QNX_SOURCE) */

	strcpy( (char *)protocol, "qcp" );
#if defined(_QNX_SOURCE)
	mname = "/dev/ser1";
#else /* !defined(_QNX_SOURCE) */
	mname = (uchar_t *)"/dev/tty00";
#endif /* defined(_QNX_SOURCE) */
	cmd_char = DEFAULT_CMD_CHAR;

#if defined(_QNX_SOURCE)
	while((opt=getopt(argc,argv,"m:l:iePEo:c:d:k:p:s:" STEAL_ALLOWED ))!=-1) {
#else /* !defined(_QNX_SOURCE) */
	while((opt=getopt(argc,argv,"m:l:iePEo:b:c:d:f:k:p:s:t:"))!=-1) {
#endif /* defined(_QNX_SOURCE) */
		switch(opt) {
			case 'm':	mname = (uchar_t *)optarg;
					break;
			case 'l':	log_file=open(optarg,O_WRONLY|O_APPEND|O_CREAT,0666);
					break;
			case 'i':	iflow = 1;
				 	break;
			case 'e':	++local_echo;
				 	break;
			case 'P':	++strip_parity;
				 	break;
			case 'E':	++qnxs;
					break;
			case 'o':	strcpy( (char *)transfer_options, optarg );
					break;
			case 'c':	cmd_char = (uchar_t)atoh( optarg );
				 	break;
			case 'd':	del_char = (uchar_t)atoh( optarg );
				 	break;
			case 'k':	kick_char = (uchar_t)atoh( optarg );
					break;
			case 'p':	pause_char = (uchar_t)atoh( optarg );
				 	break;
			case 's':	sys = (uchar_t *)optarg;
				 	break;
#ifdef STEAL_ALLOWED
			case 'S':	steal = 1;
				 	break;
#endif
#if !defined(_QNX_SOURCE)
			case 'b':	/* Baud Rate */
					if (speed_str = optarg) {
					    baud_rate = (speed_t) atoi(speed_str);
					    if (setup_baud_rate (baud_rate) == FAILURE) {
						exit (EXIT_FAILURE);
					    }
					}
					break;
			case 'f':	/* Flow Control */
					if (flow_str = optarg) {
					    if (match(flow_str,"none")) {
					        default_flow = FLOW_NONE; /* No flow control. */
					    } else if (match(flow_str,"cts_rts")) {
					        default_flow = CTS_RTS;	/* CTS/RTS flow control. */
					    } else if (match(flow_str,"xon_xoff")) {
					        default_flow = XON_XOFF; /* XON/XOFF flow control. */
					    } else {
					        fprintf (stderr,
				"%s: Flow types are none, cts_rts, xon_xoff.\n", cmdname);
						exit (EXIT_FAILURE);
					    }
					}
					break;
			case 't':	/* Parity Type ('p' & 'P' already used) */
					if (parity_str = optarg) {
					    if (match(parity_str,"even")) {
						default_parity = PARENB;
						default_data_bits = CS7;
					    } else if (match(parity_str,"odd")) {
						default_parity = (PARENB | PARODD);
						default_data_bits = CS7;
					    } else if (match(parity_str,"none")) {
						default_parity = 0;
						default_data_bits = CS8;
#if defined(_QNX_SOURCE)
					    } else if (match(parity_str,"mark")) {
						default_parity = (PARENB | PARODD | PARSTK);
						default_data_bits = CS7;
					    } else if (match(parity_str,"space")) {
						default_parity = (PARENB | PARSTK);
						default_data_bits = CS7;
#endif /* defined(_QNX_SOURCE) */
					    } else {
#if defined(_QNX_SOURCE)
						fprintf (stderr, "Valid parity settings are: even, odd, mark, space, or none.\n");
#else /* !defined(_QNX_SOURCE) */
						fprintf (stderr, "Valid parity settings are: even, odd, or none.\n");
#endif /* defined(_QNX_SOURCE) */
						exit (EXIT_FAILURE);
					    }
					}
					break;
#endif /* !defined(_QNX_SOURCE) */
			default:	exit(EXIT_FAILURE);
					/*NOTREACHED*/
		}
	}

	if (optind<argc) sys=(uchar_t *)argv[optind];

	kbd = 0;
	scr = 1;
#if defined(_QNX_SOURCE)
	mdm = open((char *)mname,O_RDWR);
#else /* !defined(_QNX_SOURCE) */
	/*
	 * O_NDELAY need to open tty line not set to CLOCAL.
	 */
	mdm = open((char *)mname,(O_RDWR | O_NDELAY));
#endif /* defined(_QNX_SOURCE) */
	if (mdm==-1) {
		fprintf(stderr, "qtalk: Unable to open device %s - %s\n",
						mname, strerror(errno) );
		exit(EXIT_FAILURE);
	}

	mname = (uchar_t *)strcpy((char *)mtbuf, ttyname(mdm));
	printf( "Using modem %s\n", mname );

#if defined(_QNX_SOURCE)
	dev_info( kbd, &kbd_entry );
	dev_info( mdm, &mdm_entry );

	if(mdm_entry.open_count > 1) {
		uchar_t nodebuf[10];
#ifndef STEAL_ALLOWED
		fprintf(stderr, "qtalk: Device %s in use.\n",mname);
		exit(EXIT_FAILURE);
#else
		if(!steal) {
			fprintf(stderr, "qtalk: Device %s in use. Use -S to try and steal.\n", mname );
			exit(EXIT_FAILURE);
		}

		for(cp=mdm_entry.tty_name+2,n=0;*cp>='0'&&*cp<='9';++cp,++n)
			nodebuf[n] = *cp;

		nodebuf[n] = '\0';

		n = spawnl(P_WAIT,	"/bin/slay", "slay",
							"-S",
							"-Q",
							"-s",
							"quit",
							"-n", nodebuf,
							"-t", cp,
							"modem",
							NULL);

		if(n < 1 || WEXITSTATUS(n) != 1) {
			fprintf(stderr, "qtalk: Unable to slay modem to steal device %s.\n", mname );
			exit(EXIT_FAILURE);
		}
#endif
	}
#endif /* defined(_QNX_SOURCE) */

#if defined(_QNX_SOURCE)

	if ( ( mproxy = qnx_proxy_attach( 0, 0, 0, -1 ) ) == -1 ) {
		fprintf(stderr, "qtalk: Unable to create modem proxy\n" );
		exit(EXIT_FAILURE);
	}

	if ( ( vmproxy = qnx_proxy_rem_attach( mdm_entry.nid, mproxy ) ) == -1 ) {
		fprintf(stderr, "qtalk: Unable to create remote modem proxy\n" );
		exit(EXIT_FAILURE);
	}

	if ( ( kproxy = qnx_proxy_attach( 0, 0, 0, -1 ) ) == -1 ) {
		fprintf(stderr, "qtalk: Unable to create keyboard proxy\n" );
		exit(EXIT_FAILURE);
	}

	if ( ( vkproxy = qnx_proxy_rem_attach( kbd_entry.nid, kproxy ) ) == -1 ) {
		fprintf(stderr, "qtalk: Unable to create remote keyboard proxy\n" );
	exit(EXIT_FAILURE);
	}

#else /* !defined(_QNX_SOURCE) */

	/*
	 * Unless debugging, don't allow anyone to steal our modem port.
	 */
	if (!debug_flag) {
	    if (ioctl (mdm, TIOCEXCL, 0) < 0) {
		perror ("TIOCEXCL");
		exit (EXIT_FAILURE);
	    }
	}

#endif /* defined(_QNX_SOURCE) */

#if defined(ultrix)
	signal( SIGINT, gotsig );
	signal( SIGQUIT, gotsig );
	signal( SIGTERM, gotsig );
#else /* !defined(ultrix) */
	signal( SIGINT, &gotsig );
	signal( SIGQUIT, &gotsig );
	signal( SIGTERM, &gotsig );
#endif /* defined(ultrix) */

	/* Set the keyboard and modem completely RAW */
#if defined(_QNX_SOURCE)
	kbd_mode = 0;
	okbd_mode = a_dev_mode( kbd, kbd_mode, _DEV_MODES);

	mdm_mode = 0;
	omdm_mode = a_dev_mode( mdm, mdm_mode, _DEV_MODES);
#endif /* defined(_QNX_SOURCE) */

	/* flush all the crud there to start with */
	(void)tcflush( mdm, TCIOFLUSH );
	(void)tcdrain( kbd );			/* Drain the output first. */
	(void)tcflush( kbd, TCIOFLUSH );

	printf( "Qtalk %s\r\n", VERSION );

	if (sys) if (dial(sys)==0) exit(EXIT_FAILURE);

#if !defined(_QNX_SOURCE)
	setup_local (kbd, &local_tty);
	setup_modem (mdm, &modem_tty);
#endif /* !defined(_QNX_SOURCE) */

	sigsetjmp( jmpbuf, 1 );	/* If we get a BREAK, send one and return here */

#if defined(_QNX_SOURCE)
	/*
	 * Arm the proxies by reading zero bytes
	 */
	a_dev_read( kbd, buf, 0, 1, 0, 0, vkproxy, 0);
	a_dev_read( mdm, buf, 0, 1, 0, 0, vmproxy, 0);

	for( ;; ) {
		pid = Receive( 0, &buf, 0 );
		if ( pid == mproxy ) {
			armed = 0;
			while(!armed) {
				n = a_dev_read( mdm, &buf[0], BUFFER_SIZE, 1, 0, 0, vmproxy, &armed);
				strncpy((char *)buf2,(char *)buf,n+1);
				buf2[n+1]=0;	
				if (n==-1) perror("dev_read(mdm...)");
				mdm_handler( &buf[0], n, &armed );
			}
		} else if ( pid == kproxy ) {
			armed = 0;
			while(!armed) {
				n = a_dev_read( kbd, &buf[0], BUFFER_SIZE, 1, 0, 0, vkproxy, &armed);
				kbd_handler( &buf[0], n );
			}
		} else Reply(pid,"\131",1);	/* reply ENOSYS */
	}
#else /* !defined(_QNX_SOURCE) */
	FD_ZERO(&readmask);
	nfds = (mdm + 1);
	while ( TRUE ) {
		/*
		 * BEWARE: The DEC OSF/1 terminal driver still has some bugs
		 *	   in it.  Occasionally, after a file transmission,
		 *	   the select will stop working for the local tty
		 *	   even though the tty characteristics are setup Ok.
		 *	   When this happens, exit 'qtalk' and restart it.
		 */
		FD_SET(kbd,&readmask), FD_SET(mdm,&readmask);
		nf = select(nfds, &readmask, NULL, NULL, (struct timeval *)0);
		if ( (nf == FAILURE) && (errno != EINTR) ) {
			 a_abrt(errno, "select", kbd);
		}
		if ( FD_ISSET(kbd, &readmask) ) {
			n = a_read(kbd, buf, a_dev_ischars(kbd));
			kbd_handler( buf, n );
		}
		if ( FD_ISSET(mdm, &readmask) ) {
			n = a_read(mdm, buf, a_dev_ischars(mdm));
			bcopy ((char *)buf, (char *)buf2, n);
			buf2[n+1] = 0;	
			mdm_handler( buf, n, &armed );
		}
	}
#endif /* defined(_QNX_SOURCE) */
	/*NOTREACHED*/
	exit (EXIT_SUCCESS);
}

#if !defined(_QNX_SOURCE)

/************************************************************************
 *									*
 *	Functions added to support non-QNX Operating Systems.		*
 *									*
 ************************************************************************/

/* Return number of characters to be read. */

int
dev_ischars(int fd)
{
	int status;
	int incnt = 0;

	if ((status = ioctl(fd, FIONREAD, (caddr_t)&incnt)) == FAILURE) {
		perror ("FIONREAD");
		return (status);
	} else {
		return (incnt);
	}
}

/* Hangup the modem by dropping DTR. */

int
tcdropline(int fd, int hsec)
{
	int status;
	unsigned int delay = (hsec / 100);

	if ((status = ioctl(fd, TIOCCDTR, 0)) < 0) {
		perror("TIOCCDTR");
		return (status);
	}
	sleep(delay);
	if ((status = ioctl(fd, TIOCSDTR, 0)) < 0) {
		perror("TIOCSDTR");
		return (status);
	}
	sleep(1);		/* Wait for DTR to reassert. */
	return (status);
}

/************************************************************************
 *									*
 * save_tty() - Save The Current Terminal Characteristics.		*
 *									*
 * Inputs:	fd = The terminal file descriptor.			*
 *		tp = The tty parameters structure.			*
 *									*
 * Return Value:							*
 *		Returns SUCCESS / FAILURE.				*
 *									*
 ************************************************************************/
int
save_tty(int fd, struct tty_params *tp)
{
	int status;

	if (tp->tty_saved == TRUE) return(SUCCESS);

#if !defined(_QNX_SOURCE)
	/*
	 * Save the line discipline.
	 */
	if ((status = ioctl (fd, TIOCGETD, &tp->tty_ldisc)) < 0) {
	    a_abrt (errno, "TIOCGETD", fd);
	}
#endif /* !defined(_QNX_SOURCE) */

	/*
	 * Save the terminal characteristics.
	 */
	if ((status = tcgetattr (fd, &tp->tty_smodes)) < 0) {
	    a_abrt (errno, "tcgetattr()", fd);
	}
	if (status == SUCCESS) tp->tty_saved = TRUE;
	tp->tty_flow = default_flow;
	tp->tty_speed = default_speed;
	tp->tty_parity = default_parity;
	tp->tty_data_bits = default_data_bits;

	return (status);
}

/************************************************************************
 *									*
 * restore_tty() - Restore Saved Terminal Characteristics.		*
 *									*
 * Inputs:	fd = The terminal file descriptor.			*
 *		tp = The tty parameters structure.			*
 *									*
 * Return Value:							*
 *		Returns SUCCESS / FAILURE / WARNING (never saved)	*
 *									*
 ************************************************************************/
int
restore_tty(int fd, struct tty_params *tp)
{
	int status;

	if (tp->tty_saved == FALSE) return (WARNING);

	/*
	 * Restore the saved terminal characteristics.
	 */
	if ((status = tcsetattr (fd, TCSANOW, &tp->tty_smodes)) < 0) {
	    a_abrt (errno, "tcsetattr()", fd);
	}

#if !defined(_QNX_SOURCE)
	/*
	 * Restore the saved line discipline.
	 */
	if (ioctl (fd, TIOCSETD, &tp->tty_ldisc) < 0) {
	    a_abrt (errno, "TIOCSETD", fd);
	}
#endif /* !defined(_QNX_SOURCE) */
	return (status);
}

/************************************************************************
 *									*
 * setup_local() - Setup local terminal characteristics.		*
 *									*
 * Inputs:	fd = The terminal file descriptor.			*
 *		tp = The tty parameters structure.			*
 *									*
 * Return Value:							*
 *		Returns SUCCESS / FAILURE.				*
 *									*
 *   --> Presumes proper line discipline has already been setup. <--	*
 *									*
 ************************************************************************/
int
setup_local(int fd, struct tty_params *tp)
{
	int status = SUCCESS;
	struct termios *tm = &tp->tty_cmodes;

	/*
	 * Save characteristics & setup initial defaults.
	 */
	if (tp->tty_saved == FALSE) {
		if ((status = save_tty (fd, tp)) == FAILURE) {
			return (status);
		}
	}
	*tm = *&tp->tty_smodes;			/* Copy saved modes. */

	/*
	 * Best Guess: Modified version of 'stty raw' mode.
	 */
	/* Hardware Control Flags */
/*	tm->c_cflag = 0;		*/

	/* Input Flags */
	tm->c_iflag &= (IXON | IXOFF | IXANY);	/* Retain flow control. */
#ifdef notdef
/*
 *	NOTE: Use 'b' command to send break sequence.  Otherwise,
 *	other characters which generate signals would have to be
 *	ignored (not really worth the trouble).
 */
	tm->c_iflag |= BRKINT;			/* map BREAK to SIGINTR */
#endif
	/* Output Flags */
	tm->c_oflag = 0;			/* No output processing. */

	/* Local Flags */
	tm->c_lflag = 0;			/* Disable echo, etc. */

	/*
	 * Setup the terminal characteristics after output is done.
	 */
	if ((status = tcsetattr (fd, TCSANOW, tm)) < 0) {
	    a_abrt (errno, "tcsetattr()", fd);
	}

	return (status);
}

/************************************************************************
 *									*
 * setup_modem() - Setup modem terminal characteristics.		*
 *									*
 * Inputs:	fd = The terminal file descriptor.			*
 *		tp = The tty parameters structure.			*
 *									*
 * Return Value:							*
 *		Returns SUCCESS / FAILURE.				*
 *									*
 *   --> Presumes proper line discipline has already been setup. <--	*
 *									*
 ************************************************************************/
int
setup_modem(int fd, struct tty_params *tp)
{
	int status = SUCCESS;
	struct termios *tm = &tp->tty_cmodes;

	/*
	 * Save characteristics & setup initial defaults.
	 */
	if (tp->tty_saved == FALSE) {
		if ((status = save_tty (fd, tp)) == FAILURE) {
			return (status);
		}
	}
	*tm = *&tp->tty_smodes;			/* Copy saved modes. */

	/*
	 * Initialize all the tty mode flags/speed.
	 */
	tm->c_cflag = tm->c_iflag = tm->c_oflag = tm->c_lflag = 0;

	/*
	 * Ensure these are set to normal default values.
	 */
	tm->c_cc[VMIN] = 1;
	tm->c_cc[VTIME] = 0;

	/*
	 * Remember, tty *must* be set to CLOCAL or the terminal driver
	 * won't let us talk to the modem (reads hang, writes get EIO).
	 */
	if (modem_flag) {
	    tm->c_cflag |= HUPCL;	/* Hangup on last close. */
	} else {
	    tm->c_cflag |= CLOCAL;	/* Ignore modem signals. */
	}

	tm->c_cflag |= (tp->tty_data_bits | tp->tty_parity | CREAD);
#if defined(ultrix)
	tm->c_cflag |= ( (tp->tty_speed << 16) | tp->tty_speed);
#else /* !defined(ultrix) */
	tm->c_ispeed = tm->c_ospeed = tp->tty_speed;
#endif /* defined(ultrix) */

	/*
	 * Send two stop bits for slower speeds (is this right?).
	 */
	if (tp->tty_speed < B300) {
	    tm->c_cflag |= CSTOPB;
	}
	/*
	 * Set the desired flow control.
	 */
#if defined(ultrix)
	if (tp->tty_flow == XON_XOFF) {
	    tm->c_iflag |= (IXON | IXOFF);	/* XON/XOFF Flow control. */
	}
#elif defined(_OSF_SOURCE)
	if (tp->tty_flow == CTS_RTS) {
	    tm->c_cflag |= CRTSCTS;		/* CTS/RTS flow control. */
	} else if (tp->tty_flow == XON_XOFF) {
	    tm->c_iflag |= (IXON | IXOFF);	/* XON/XOFF Flow control. */
	}
#elif defined(_QNX_SOURCE)
	tm->c_lflag |= IEXTEN;			/* QNX POSIX extensions. */
	if (tp->tty_flow == CTS_RTS) {
	    tm->c_lflag |= (IHFLOW | OHFLOW);	/* CTS/RTS flow control. */
	} else if (tp->tty_flow == XON_XOFF) {
	    tm->c_iflag |= (IXON | IXOFF);	/* XON/XOFF Flow control. */
	}
#endif /* defined(ultrix) */

	/*
	 * If 7 bit data, or parity is being enabled, enable stripping
	 * of the eighth data bit, otherwise the driver passes it back.
	 * We don't test generation of correct parity, driver does this.
	 */
	if ( (tp->tty_data_bits == CS7) || (tp->tty_parity & PARENB) ) {
	    tm->c_iflag |= ISTRIP;		/* Strip the 8th data bit */
	    /*
	     * For parity, enable input parity checking.
	     */
	    if (tp->tty_parity & PARENB) {
		tm->c_iflag |= INPCK;		/* Check input parity.	*/
	    }
	}

	/*
	 * Setup the terminal characteristics after output is done.
	 */
	if ((status = tcsetattr (fd, TCSANOW, tm)) < 0) {
	    a_abrt (errno, "tcsetattr()", fd);
	}

	return (status);
}

/************************************************************************
 *									*
 * setup_baud_rate() - Setup the baud rate code.			*
 *									*
 *		This function takes a specified baud rate (i.e. 9600)	*
 *		and maps it to the baud rate code we must specify to	*
 *		the Unix terminal driver to set that baud rate.		*
 *									*
 * Inputs:	baud = The real baud rate.				*
 *									*
 * Return Value:							*
 *		Returns SUCCESS / FAILURE = Valid Speed/Invalid Speed.	*
 *									*
 ************************************************************************/
int
setup_baud_rate (baud)
register speed_t baud;
{
    register int i;
    register struct tty_baud_rate *tsp;

    for (tsp = baud_rate_tbl, i = 0; i < num_baud_rates; i++, tsp++) {
	if (baud == tsp->usr_speed) {
	    default_speed = tsp->tty_speed;	/* Save baud rate code. */
	    return (SUCCESS);			/* Return success status. */
	}
    }
    fprintf (stderr, "Baud rate '%d' is invalid, valid entrys are:\n", baud);
    for (tsp = baud_rate_tbl, i = 0; i < num_baud_rates; i++, tsp++) {
	if ( (i % 6) == 0) fprintf (stderr, "\n");
	fprintf (stderr, "%10d", tsp->usr_speed);
    }
    fprintf (stderr, "\n");
    return (FAILURE);
}

#endif /* !defined(_QNX_SOURCE) */
